#include "cduvmap.h"
#include <stdio.h>

#define UV_EPSILON (0.000005f)

//---------------------------------------------------------------------------------------
CDUVMap::CDUVMap()
{
   mUseSS = false;
   mSpacePoles = true;
}
//---------------------------------------------------------------------------------------
CDUVMap::~CDUVMap()
{
}
//---------------------------------------------------------------------------------------
void CDUVMap::GetSelectionBounds( List *surfs, Vertex *vmin, Vertex *vmax, Vertex *vcnt )
{

   if( surfs == NULL )
   {
      return;
   }

   Vertex tvmin, tvmax;
   Vertex *tv = (( SVertex * )((( Surface * )( surfs->data ))->vertlist->data ))->v;
   if( mUseSS )
   {
      tv = mSSMap.GetSSVertex( tv );
   }
   tvmin.x = tv->x;
   tvmin.y = tv->y;
   tvmin.z = tv->z;
   tvmax.x = tv->x;
   tvmax.y = tv->y;
   tvmax.z = tv->z;

   for( List *ps = surfs; ps != NULL; ps = ps->next )
   {
      for( List *psv = (( Surface * )( ps->data ))->vertlist; psv != NULL; psv = psv->next )
      {
         Vertex *v = (( SVertex * )( psv->data ))->v;
         if( mUseSS )
         {
            v = mSSMap.GetSSVertex( v );
         }
         tvmin.x = min( tvmin.x, v->x );
         tvmin.y = min( tvmin.y, v->y );
         tvmin.z = min( tvmin.z, v->z );
         tvmax.x = max( tvmax.x, v->x );
         tvmax.y = max( tvmax.y, v->y );
         tvmax.z = max( tvmax.z, v->z );
      }
   }

   if( vcnt != NULL )
   {
      vcnt->x = ( tvmin.x + tvmax.x ) * 0.5f;
      vcnt->y = ( tvmin.y + tvmax.y ) * 0.5f;
      vcnt->z = ( tvmin.z + tvmax.z ) * 0.5f;
   }
   if( vmin != NULL )
   {
      vmin->x = tvmin.x;
      vmin->y = tvmin.y;
      vmin->z = tvmin.z;
   }
   if( vmax != NULL )
   {
      vmax->x = tvmax.x;
      vmax->y = tvmax.y;
      vmax->z = tvmax.z;
   }

}
//---------------------------------------------------------------------------------------
void CDUVMap::FixUVWrap( Surface *s )
{

   //===============================================================
   // Courtesy Andy Colebourne, with edits to get min/max tu:
   // check for overlaps in spherical mapping
   int flag = 0;
   SVertex *jv, *kv;

   for (int j=0; j < s->numvert; j++)
   {
      for (int k=0; k < s->numvert; k++)
      {
         jv = surface_get_svertex_at(s, j);
         kv = surface_get_svertex_at(s, k);

         if (0.5 <= fabs(jv->tx - kv->tx))
            flag |= 1;
      }
   }

   float minu = 100.0f, maxu = -100.0f;
   if (flag == 1)
   {
      for (int k=0; k < s->numvert; k++)
      {
         SVertex *jv = surface_get_svertex_at(s, k);
         if (jv->tx < 0.5)
            jv->tx += 1.0;
         minu = min( jv->tx, minu );
         maxu = max( jv->tx, maxu );
      } 
   }
   //===============================================================

   // Pull texture coordinates back if over 1/2 of the horizontal bounds is past 1.0
   if( 1.0f - minu < maxu - 1.0f )
   {
      for (int k=0; k < s->numvert; k++)
      {
         SVertex *jv = surface_get_svertex_at(s, k);
         jv->tx -= 1.0f;
      } 
   }

}
//---------------------------------------------------------------------------------------
void CDUVMap::SpacePolarCoordinates( List *surfs, int axis, Vertex *vcenter )
{
   for( List *ps = surfs; ps != NULL; ps = ps->next )
   {
      Surface *s = ( Surface * )( ps->data );
      for( List *psv = s->vertlist; psv != NULL; psv = psv->next )
      {
         SVertex *sv = ( SVertex * )( psv->data );
         Vertex *v = sv->v;
         bool polar = false;
         switch( axis )
         {
         case 0:
            // X axis
            if( fabs( vcenter->y - v->y ) < UV_EPSILON &&
                fabs( vcenter->z - v->z ) < UV_EPSILON )
            {
               polar = true;
            }
            break;
         case 2:
            // Z axis
            if( fabs( vcenter->y - v->y ) < UV_EPSILON &&
                fabs( vcenter->x - v->x ) < UV_EPSILON )
            {
               polar = true;
            }
            break;
         default:
            // Y axis
            if( fabs( vcenter->x - v->x ) < UV_EPSILON &&
                fabs( vcenter->z - v->z ) < UV_EPSILON )
            {
               polar = true;
            }
            break;
         }
         if( polar )
         {
            // Move the horizontal UV to the unweighted center
            float tumin, tumax;
            bool first = true;
            for( List *psv2 = s->vertlist; psv2 != NULL; psv2 = psv2->next )
            {
               SVertex *sv2 = ( SVertex * )( psv2->data );
               if( sv == sv2 )
               {
                  continue;
               }
               if( first )
               {
                  tumin = sv2->tx;
                  tumax = sv2->tx;
                  first = false;
               }
               else
               {
                  tumin = min( tumin, sv2->tx );
                  tumax = max( tumax, sv2->tx );
               }
            }
            // Ensure we read at least one entry (!first)
            if( !first )
            {
               sv->tx = tumin + (( tumax - tumin ) * 0.5f );
            }
         }
      }
   }
}
//---------------------------------------------------------------------------------------
void CDUVMap::ApplyPlanar( List *surfs, int plane )
{

   Vertex vmin, vmax;
   GetSelectionBounds( surfs, &vmin, &vmax, NULL );
   Vertex size;
   size.x = vmax.x - vmin.x;
   size.y = vmax.y - vmin.y;
   size.z = vmax.z - vmin.z;

   // Apply new coordinates to SVertices
   for( List *ps = surfs; ps != NULL; ps = ps->next )
   {
      Surface *s = ( Surface * )( ps->data );
      for( List *psv = s->vertlist; psv != NULL; psv = psv->next )
      {
         SVertex *sv = ( SVertex * )( psv->data );
         Vertex *v = sv->v;
         if( mUseSS )
         {
            v = mSSMap.GetSSVertex( v );
         }
         // Calculate
         float thisx, thisy;
         switch( plane )
         {
         case 3:
         case 4:
            // Left/right
            thisx = ( v->z - vmin.z ) / size.z;
            thisy = ( v->y - vmin.y ) / size.y;
            if( plane == 4 )
            {
               thisx = 1.0f - thisx;
            }
            break;
         case 5:
         case 6:
            // Top/bottom
            thisx = ( v->x - vmin.x ) / size.x;
            thisy = -(( v->z - vmin.z ) / size.z );
            if( plane == 6 )
            {
               thisx = 1.0f - thisx;
            }
            break;
         default:
            // Front/back
            thisx = ( v->x - vmin.x ) / size.x;
            thisy = ( v->y - vmin.y ) / size.y;
            if( plane == 2 )
            {
               thisx = 1.0f - thisx;
            }
            break;
         }
         sv->tx = thisx;
         sv->ty = thisy;
      }
   }

}
//---------------------------------------------------------------------------------------
void CDUVMap::ApplyBox( List *surfs )
{

   Vertex vmin, vmax;
   GetSelectionBounds( surfs, &vmin, &vmax, NULL );
   Vertex size;
   size.x = vmax.x - vmin.x;
   size.y = vmax.y - vmin.y;
   size.z = vmax.z - vmin.z;

   // Apply new coordinates to SVertices
   for( List *ps = surfs; ps != NULL; ps = ps->next )
   {
      Surface *s = ( Surface * )( ps->data );
      // Get this surface's plane
      int plane = C3d::GetDominantPlaneFromNormal( s->normal );
      for( List *psv = s->vertlist; psv != NULL; psv = psv->next )
      {
         SVertex *sv = ( SVertex * )( psv->data );
         Vertex *v = sv->v;
         if( mUseSS )
         {
            v = mSSMap.GetSSVertex( v );
         }
         // Calculate
         float thisx, thisy;
         switch( plane )
         {
         case C3D_PLANE_X:
            // Left/right
            thisx = ( v->z - vmin.z ) / size.z;
            thisy = ( v->y - vmin.y ) / size.y;
            if( s->normal.x > 0 )
            {
               thisx = 1.0f - thisx;
            }
            break;
         case C3D_PLANE_Y:
            // Top/bottom
            thisx = ( v->x - vmin.x ) / size.x;
            thisy = -(( v->z - vmin.z ) / size.z );
            if( s->normal.y < 0 )
            {
               thisx = 1.0f - thisx;
            }
            break;
         default:
            // Front/back
            thisx = ( v->x - vmin.x ) / size.x;
            thisy = ( v->y - vmin.y ) / size.y;
            if( s->normal.z < 0 )
            {
               thisx = 1.0f - thisx;
            }
            break;
         }
         sv->tx = thisx;
         sv->ty = thisy;
      }
   }

}
//---------------------------------------------------------------------------------------
void CDUVMap::ApplySpherical( List *surfs, int axis )
{

   Vertex vcenter;
   GetSelectionBounds( surfs, NULL, NULL, &vcenter );

   Vertex vray;

   Vertex north;
   Vertex equator;
   Vertex northequatorx;
   switch( axis )
   {
   case 0:
      // X axis
      north.x = 1;
      north.y = 0;
      north.z = 0;
      equator.x = 0;
      equator.y = 0;
      equator.z = 1;
      break;
   case 2:
      // Z axis
      north.x = 0;
      north.y = 0;
      north.z = 1;
      equator.x = 0;
      equator.y = 1;
      equator.z = 0;
      break;
   default:
      // X axis
      north.x = 0;
      north.y = 1;
      north.z = 0;
      equator.x = 0;
      equator.y = 0;
      equator.z = 1;
      break;
   }
   CROSSPRODUCT( &north, &equator, &northequatorx );

   // Apply new coordinates to SVertices
   for( List *ps = surfs; ps != NULL; ps = ps->next )
   {
      Surface *s = ( Surface * )( ps->data );
      for( List *psv = s->vertlist; psv != NULL; psv = psv->next )
      {
         SVertex *sv = ( SVertex * )( psv->data );
         Vertex *v = sv->v;
         if( mUseSS )
         {
            v = mSSMap.GetSSVertex( v );
         }
         // Cast a ray from the center through the current vertex
         vray.x = v->x - vcenter.x;
         vray.y = v->y - vcenter.y;
         vray.z = v->z - vcenter.z;
         C3d::Normalize( vray );
         // Get the angle between "north" and the ray cast through the center
         float phi = acos( DOTPRODUCT( &north, &vray ));
         // Apply the V coordinate
         sv->ty = 1.0f - ( phi / M_PI );
         if( axis == 0 && ( vray.x == 1.0f || vray.x == -1.0f ))
         {
            sv->tx = 0.5f;
         }
         else if( axis == 1 && ( vray.y == 1.0f || vray.y == -1.0f ))
         {
            sv->tx = 0.5f;
         }
         else if( axis == 2 && ( vray.z == 1.0f || vray.z == -1.0f ))
         {
            sv->tx = 0.5f;
         }
         else
         {
            sv->tx = acos( max( min( DOTPRODUCT( &equator, &vray ) / sin( phi ), 1.0f ), -1.0f )) / ( 2.0f * M_PI ) + 0.5f;
            sv->tx = ( DOTPRODUCT( &northequatorx, &vray ) > 0 ) ? sv->tx : 1.0f - sv->tx;
         }
         if( axis == 2 )
         {
            sv->tx = 1.0f - sv->tx;
            sv->ty = -( sv->ty );
         }
      }
      // Adjust UV coordinates for the surface
      FixUVWrap( s );
   }

   if( mSpacePoles )
   {
      SpacePolarCoordinates( surfs, axis, &vcenter );
   }

}
//---------------------------------------------------------------------------------------
void CDUVMap::ApplyCylindrical( List *surfs, int axis )
{

   // This is basically the same as ApplySpherical, but the axis coordinate is applied based
   // on the height of the selection

   Vertex vcenter, vmin, vmax;
   GetSelectionBounds( surfs, &vmin, &vmax, &vcenter );

   Vertex vray;

   float selheight;
   Vertex north;
   Vertex equator;
   Vertex northequatorx;
   switch( axis )
   {
   case 0:
      // X axis
      north.x = 1;
      north.y = 0;
      north.z = 0;
      equator.x = 0;
      equator.y = 0;
      equator.z = 1;
      selheight = ( vmax.x - vmin.x );
      break;
   case 2:
      // Z axis
      north.x = 0;
      north.y = 0;
      north.z = 1;
      equator.x = 0;
      equator.y = 1;
      equator.z = 0;
      selheight = ( vmax.z - vmin.z );
      break;
   default:
      // X axis
      north.x = 0;
      north.y = 1;
      north.z = 0;
      equator.x = 0;
      equator.y = 0;
      equator.z = 1;
      selheight = ( vmax.y - vmin.y );
      break;
   }
   CROSSPRODUCT( &north, &equator, &northequatorx );

   // Apply new coordinates to SVertices
   for( List *ps = surfs; ps != NULL; ps = ps->next )
   {
      Surface *s = ( Surface * )( ps->data );
      for( List *psv = s->vertlist; psv != NULL; psv = psv->next )
      {
         SVertex *sv = ( SVertex * )( psv->data );
         Vertex *v = sv->v;
         if( mUseSS )
         {
            v = mSSMap.GetSSVertex( v );
         }
         // Cast a ray from the center through the current vertex
         vray.x = v->x - vcenter.x;
         vray.y = v->y - vcenter.y;
         vray.z = v->z - vcenter.z;
         C3d::Normalize( vray );
         // Get the angle between "north" and the ray cast through the center
         float phi = acos( DOTPRODUCT( &north, &vray ));
         if( axis == 0 && ( vray.x == 1.0f || vray.x == -1.0f ))
         {
            sv->tx = 0.5f;
         }
         else if( axis == 1 && ( vray.y == 1.0f || vray.y == -1.0f ))
         {
            sv->tx = 0.5f;
         }
         else if( axis == 2 && ( vray.z == 1.0f || vray.z == -1.0f ))
         {
            sv->tx = 0.5f;
         }
         else
         {
            sv->tx = acos( max( min( DOTPRODUCT( &equator, &vray ) / sin( phi ), 1.0f ), -1.0f )) / ( 2.0f * M_PI ) + 0.5f;
            sv->tx = ( DOTPRODUCT( &northequatorx, &vray ) > 0 ) ? sv->tx : 1.0f - sv->tx;
         }
         // Apply the height coordinate
         switch( axis )
         {
         case 0:
            sv->ty = (( v->x - vmin.x ) / selheight );
            break;
         case 2:
            sv->tx = 1.0f - sv->tx;
            sv->ty = -(( v->z - vmin.z ) / selheight );
            break;
         default:
            sv->ty = (( v->y - vmin.y ) / selheight );
            break;
         }
      }
      // Adjust UV coordinates for the surface
      FixUVWrap( s );
   }

   if( mSpacePoles )
   {
      SpacePolarCoordinates( surfs, axis, &vcenter );
   }

}
//---------------------------------------------------------------------------------------
void CDUVMap::ApplyCylindricalCap( List *surfs, int axis )
{

   // This is basically the same as ApplyCylindrical, but box mapping is applied
   // to faces with up/down normals

   Vertex vcenter, vmin, vmax;
   GetSelectionBounds( surfs, &vmin, &vmax, &vcenter );
   Vertex size;
   size.x = vmax.x - vmin.x;
   size.y = vmax.y - vmin.y;
   size.z = vmax.z - vmin.z;

   Vertex vray;

   float selheight;
   Vertex north;
   Vertex equator;
   Vertex northequatorx;
   switch( axis )
   {
   case 0:
      // X axis
      north.x = 1;
      north.y = 0;
      north.z = 0;
      equator.x = 0;
      equator.y = 0;
      equator.z = 1;
      selheight = ( vmax.x - vmin.x );
      break;
   case 2:
      // Z axis
      north.x = 0;
      north.y = 0;
      north.z = 1;
      equator.x = 0;
      equator.y = 1;
      equator.z = 0;
      selheight = ( vmax.z - vmin.z );
      break;
   default:
      // X axis
      north.x = 0;
      north.y = 1;
      north.z = 0;
      equator.x = 0;
      equator.y = 0;
      equator.z = 1;
      selheight = ( vmax.y - vmin.y );
      break;
   }
   CROSSPRODUCT( &north, &equator, &northequatorx );

   // Apply new coordinates to SVertices
   for( List *ps = surfs; ps != NULL; ps = ps->next )
   {
      Surface *s = ( Surface * )( ps->data );
      // Get this surface's plane
      int plane = C3d::GetDominantPlaneFromNormal( s->normal );
      bool fixuvs = false;
      for( List *psv = s->vertlist; psv != NULL; psv = psv->next )
      {
         SVertex *sv = ( SVertex * )( psv->data );
         Vertex *v = sv->v;
         if( mUseSS )
         {
            v = mSSMap.GetSSVertex( v );
         }

         if( plane == C3D_PLANE_Y && axis == 1 )
         {
            // Top/bottom
            sv->tx = ( v->x - vmin.x ) / size.x;
            sv->ty = -(( v->z - vmin.z ) / size.z );
            if( s->normal.y < 0 )
            {
               sv->ty = 1.0f - sv->ty;
            }
         }
         else if( plane == C3D_PLANE_X && axis == 0 )
         {
            // Left/right
            sv->tx = -(( v->y - vmin.y ) / size.y );
            sv->ty = -(( v->z - vmin.z ) / size.z );
            if( s->normal.x < 0 )
            {
               sv->ty = -( ( -1.0f ) + sv->ty );
            }
         }
         else if( plane == C3D_PLANE_Z && axis == 2 )
         {
            // Front/back
            sv->tx = ( v->x - vmin.x ) / size.x;
            sv->ty = (( v->y - vmin.y ) / size.y );
            if( s->normal.z < 0 )
            {
               sv->ty = 1.0f - sv->ty;
            }
         }
         else
         {
            fixuvs = true;
            // Cast a ray from the center through the current vertex
            vray.x = v->x - vcenter.x;
            vray.y = v->y - vcenter.y;
            vray.z = v->z - vcenter.z;
            C3d::Normalize( vray );
            // Get the angle between "north" and the ray cast through the center
            float phi = acos( DOTPRODUCT( &north, &vray ));
            //sv->ty = 1.0f - ( phi / M_PI );
            if( axis == 0 && ( vray.x == 1.0f || vray.x == -1.0f ))
            {
               sv->tx = 0.5f;
            }
            else if( axis == 1 && ( vray.y == 1.0f || vray.y == -1.0f ))
            {
               sv->tx = 0.5f;
            }
            else if( axis == 2 && ( vray.z == 1.0f || vray.z == -1.0f ))
            {
               sv->tx = 0.5f;
            }
            else
            {
               sv->tx = acos( max( min( DOTPRODUCT( &equator, &vray ) / sin( phi ), 1.0f ), -1.0f )) / ( 2.0f * M_PI ) + 0.5f;
               sv->tx = ( DOTPRODUCT( &northequatorx, &vray ) > 0 ) ? sv->tx : 1.0f - sv->tx;
            }
            // Apply the height coordinate
            switch( axis )
            {
            case 0:
               sv->ty = (( v->x - vmin.x ) / selheight );
               break;
            case 2:
               sv->tx = 1.0f - sv->tx;
               //sv->tx = -( sv->tx );
               sv->ty = -(( v->z - vmin.z ) / selheight );
               break;
            default:
               sv->ty = (( v->y - vmin.y ) / selheight );
               break;
            }
         }
      }
      // Adjust UV coordinates for the surface
      if( fixuvs )
      {
         FixUVWrap( s );
      }
   }

}
//---------------------------------------------------------------------------------------
void CDUVMap::ApplyGlobal( List *surfs, float scalex, float scaley, float offsetx, float offsety, float rotation )
{

   // See if we're rotating (rotation != 0)
   bool rotating = ( fabs( rotation ) > 0.00001f );

   // Create sin() and cos() vars of our desired rotation (so we don't repeat
   // the calculation for each vertex)
   float rotcos, rotsin;
   if( rotating )
   {
      rotcos = cos( DEGTORAD( rotation ));
      rotsin = sin( DEGTORAD( rotation ));
   }

   // Offset/scale are provided as a "whole number percentage", so divide by 100
   offsetx /= 100.0f;
   offsety /= 100.0f;
   scalex /= 100.0f;
   scaley /= 100.0f;

   // Calculate coordinates for each selected surface
   for( List *p = surfs; p != NULL; p = p->next )
   {

      // Get a pointer to the surface
      Surface *surf = ( Surface * )( p->data );

      // Get texture size
      int texw, texh, texd;
      ACObject *obj = object_of_surface( surf );
      if( ac_object_has_texture( obj ))
      {
         ac_image_get_dim( texture_id_to_image( ac_object_get_texture_index( obj )),
                           &texw, &texh, &texd );
      }
      else
      {
         texw = 1;
         texh = 1;
      }
      // Set our width/height to 1.0 for the purposes of texture mapping.
      float ftexw, ftexh;
      if( texw == texh )
      {
         ftexw = 1.0f;
         ftexh = 1.0f;
      }
      else if( texw > texh )
      {
         ftexh = ( float )texh / ( float )texw;
         ftexw = 1.0f;
      }
      else
      {
         ftexw = ( float )texw / ( float )texh;
         ftexh = 1.0f;
      }

      Point3 *norm = &surf->normal;

      // Reports whether we need to normalize X/Y texture coordinates
      bool normx = true;
      bool normy = true;

      // Get the dominant plane
      int plane = C3d::GetDominantPlaneFromNormal( *norm );

      for( List *pv = surf->vertlist; pv != NULL; pv = pv->next )
      {

         // Get a pointer to the surface
         SVertex *vert = ( SVertex * )( pv->data );
         Vertex *vpos = vert->v;
         if( mUseSS )
         {
            vpos = mSSMap.GetSSVertex( vpos );
         }

         // Calculate coordinates based on the texture axis (same as the surface normal).
         // Use the dominant normal as the texture axis.
         // In this routine, we're reversing the image's "X" orientation for
         //    back-facing surfaces.  This way, the texture is always "left-to-right",
         //    no matter which side the normal faces.
         float tvertx, tverty;
         switch( plane )
         {
         case C3D_PLANE_Z:
            if( norm->z < 0.0 )
            {
               tvertx = - vpos->x;
               tverty =   vpos->y;
            }
            else
            {
               tvertx = vpos->x;
               tverty = vpos->y;
            }
            break;
         case C3D_PLANE_Y:
            // Y axis
            if( norm->y < 0.0 )
            {
               tvertx = vpos->x;
               tverty = vpos->z;
            }
            else
            {
               tvertx = vpos->x;
               tverty = - vpos->z;
            }
            break;
         default:
            // X axis
            if( norm->x < 0.0 )
            {
               tvertx = vpos->z;
               tverty = vpos->y;
            }
            else
            {
               tvertx = - vpos->z;
               tverty =   vpos->y;
            }
            break;
         }

         // tvertx and tverty contains a copy of the significant vertices.
         // Rotate them from the origin if the user wants rotation.
         if( rotating )
         {
            float tvx = tvertx;  // keep an uncorrupted copy
            tvertx = ( tvertx * rotcos ) - ( tverty * rotsin );
            tverty = ( tvx    * rotsin ) + ( tverty * rotcos );
         }

         // Finally, apply scaling and offset.
         vert->tx = tvertx / ( ftexw * scalex ) - offsetx;
         vert->ty = tverty / ( ftexh * scaley ) - offsety;

         // If we find any texture coordinates in -1..1 range, we can assume that
         // the coordinates are as "normalized" as they're going to get.
         if( vert->tx < 1 && vert->tx > -1 )
         {
            normx = false;
         }
         if( vert->ty < 1 && vert->ty > -1 )
         {
            normy = false;
         }

      }

      // Done generating coordinates for this face.
      // Now, normalize the generated texture coordinates by bringing the smallest
      // value to between -1 and 1.
      if( normx || normy )
      {

         float absx, absy, x, y;

         // Get the coordinates closest to zero for the first vertex
         // X coordinate closest to zero
         absx = 0.0;
         x = (( SVertex * )surf->vertlist->data )->tx;
         if( x > 1.0 )
         {
            absx = floor( x );
         }
         else
         {
            absx = ceil( x );
         }
         // Y coordinate closest to zero
         absy = 0.0;
         y = (( SVertex * )surf->vertlist->data )->ty;
         if( y > 1.0 )
         {
            absy = floor( y );
         }
         else
         {
            absy = ceil( y );
         }

         // Get the coordinates closest to zero for the rest of the vertices
         for( List *pv = surf->vertlist; pv != NULL; pv = pv->next )
         {
            SVertex *vert = ( SVertex * )( pv->data );
            // X coordinates
            if( normx )
            {
               x = vert->tx;
               if( fabs( x ) < fabs( absx ))
               {
                  if( x > 1.0 )
                  {
                     absx = floor( x );
                  }
                  else
                  {
                     absx = ceil( x );
                  }
               }
            }
            // Y coordinates
            if( normy )
            {
               y = vert->ty;
               if( fabs( y ) < fabs( absy ))
               {
                  if( y > 1.0 )
                  {
                     absy = floor( y );
                  }
                  else
                  {
                     absy = ceil( y );
                  }
               }
            }
         }

         // Normalize all of the vertices in the surface by subtracting
         // the value closest to zero from all texture coordinates in the face.
         for( List *pv = surf->vertlist; pv != NULL; pv = pv->next )
         {
            SVertex *vert = ( SVertex * )( pv->data );
            vert->tx = vert->tx - absx;
            vert->ty = vert->ty - absy;
         }

      }

   }

}
//---------------------------------------------------------------------------------------
